@@ -1,6 +1,6 @@
 package Apache::AuthCookie;
-BEGIN {
-  $Apache::AuthCookie::VERSION = '3.18';
+{
+  $Apache::AuthCookie::VERSION = '3.20';
 }
 
 # ABSTRACT: Perl Authentication and Authorization via cookies
@@ -9,22 +9,24 @@ use strict;
 
 use Carp;
 use mod_perl qw(1.07 StackedHandlers MethodHandlers Authen Authz);
-use Apache::Constants qw(:common M_GET FORBIDDEN REDIRECT);
+use Apache::Constants qw(:common M_GET FORBIDDEN OK REDIRECT);
+use Apache::AuthCookie::Params;
 use Apache::AuthCookie::Util;
+use Apache::AuthCookie::Autobox;
 use Apache::Util qw(escape_uri);
 
 sub recognize_user ($$) {
     my ($self, $r) = @_;
 
     # only check if user is not already set
-    return DECLINED if $r->connection->user;
+    return DECLINED unless $r->connection->user->is_blank;
 
     my $debug = $r->dir_config("AuthCookieDebug") || 0;
     my ($auth_type, $auth_name) = ($r->auth_type, $r->auth_name);
 
-    return DECLINED unless $auth_type && $auth_name;
+    return DECLINED if $auth_type->is_blank or $auth_name->is_blank;
 
-    return DECLINED unless $r->header_in('Cookie');
+    return DECLINED if $r->header_in('Cookie')->is_blank;
 
     my $cookie_name = $self->cookie_name($r);
 
@@ -33,7 +35,7 @@ sub recognize_user ($$) {
     return DECLINED unless $cookie;
 
     my ($user, @args) = $auth_type->authen_ses_key($r, $cookie);
-    if ($user and scalar @args == 0) {
+    if (!$user->is_blank and scalar @args == 0) {
         $r->log_error("user is $user") if $debug >= 2;
 
         # if SessionTimeout is on, send new cookie with new Expires.
@@ -47,7 +49,7 @@ sub recognize_user ($$) {
         return $auth_type->custom_errors($r, $user, @args);
     }
 
-    return OK;
+    return $user->is_blank ? DECLINED : OK;
 }
 
 sub cookie_name {
@@ -93,13 +95,13 @@ sub remove_cookie {
     );
 
     $r->err_headers_out->add("Set-Cookie" => "$str");
-    $r->log_error("removed_cookie " . $r->err_headers_out->get("Set-Cookie"))
-        if $debug >= 2;
+
+    $r->log_error("removed cookie $cookie_name") if $debug >= 2;
 }
 
 # convert current request to GET
 sub _convert_to_get {
-    my ($self, $r, $args) = @_;
+    my ($self, $r) = @_;
 
     return unless $r->method eq 'POST';
 
@@ -107,17 +109,16 @@ sub _convert_to_get {
 
     $r->log_error("Converting POST -> GET") if $debug >= 2;
 
+    my $args = $self->params($r);
+
     my @pairs = ();
-    while (my ($name, $value) = each %$args) {
 
+    for my $name ($args->param) {
         # we dont want to copy login data, only extra data
-        next
-            if $name eq 'destination'
-                or $name =~ /^credential_\d+$/;
+        next if $name eq 'destination'
+             or $name =~ /^credential_\d+$/;
 
-        $value = '' unless defined $value;
-
-        for my $v (split /\0/, $value) {
+        for my $v ($args->param($name)) {
             push @pairs, escape_uri($name) . '=' . escape_uri($v);
         }
     }
@@ -129,23 +130,10 @@ sub _convert_to_get {
     $r->headers_in->unset('Content-Length');
 }
 
-sub _get_form_data {
+sub params {
     my ($self, $r) = @_;
 
-    my @pairs = $r->method eq 'POST' ? $r->content : $r->args;
-
-    my %vars = ();
-
-    while (my ($name, $value) = splice @pairs, 0, 2) {
-        unless (defined $vars{$name}) {
-            $vars{$name} = $value;
-        }
-        else {
-            $vars{$name} .= "\0$value";
-        }
-    }
-
-    return %vars;
+    return Apache::AuthCookie::Params->new($r);
 }
 
 sub login ($$) {
@@ -153,11 +141,12 @@ sub login ($$) {
     my $debug = $r->dir_config("AuthCookieDebug") || 0;
 
     my ($auth_type, $auth_name) = ($r->auth_type, $r->auth_name);
-    my %args = $self->_get_form_data($r);
 
-    $self->_convert_to_get($r, \%args) if $r->method eq 'POST';
+    my $params = $self->params($r);
 
-    unless (exists $args{'destination'}) {
+    $self->_convert_to_get($r) if $r->method eq 'POST';
+
+    unless (defined $params->param('destination')) {
         $r->log_error("No key 'destination' found in form data");
         $r->subprocess_env('AuthCookieReason', 'no_cookie');
         return $auth_type->login_form;
@@ -165,10 +154,11 @@ sub login ($$) {
 
     # Get the credentials from the data posted by the client
     my @credentials;
-    for (my $i = 0 ; exists $args{"credential_$i"} ; $i++) {
+    for (my $i = 0 ; defined $params->param("credential_$i") ; $i++) {
         my $key = "credential_$i";
-        $r->log_error("$key $args{$key}") if $debug >= 2;
-        push @credentials, $args{$key};
+        my $val = $params->param("credential_$i");
+        $r->log_error("$key $val") if $debug >= 2;
+        push @credentials, $val;
     }
 
     # save creds in pnotes in case login form script wants to use them.
@@ -179,7 +169,7 @@ sub login ($$) {
     unless ($ses_key) {
         $r->log_error("Bad credentials") if $debug >= 2;
         $r->subprocess_env('AuthCookieReason', 'bad_credentials');
-        $r->uri($args{'destination'});
+        $r->uri($self->untaint_destination($params->param('destination')));
         return $auth_type->login_form;
     }
 
@@ -197,7 +187,7 @@ sub login ($$) {
     $self->handle_cache;
 
     $r->header_out(
-        "Location" => $self->untaint_destination($args{'destination'}));
+        "Location" => $self->untaint_destination($params->param('destination')));
 
     return REDIRECT;
 }
@@ -272,7 +262,7 @@ sub authenticate ($$) {
         my ($auth_user, @args) =
             $auth_type->authen_ses_key($r, $ses_key_cookie);
 
-        if ($auth_user and scalar @args == 0) {
+        if (!$auth_user->is_blank and scalar @args == 0) {
 
             # We have a valid session key, so we return with an OK value.
             # Tell the rest of Apache what the authentication method and
@@ -317,9 +307,7 @@ sub login_form {
     my $r = Apache->request or die "no request";
     my $auth_name = $r->auth_name;
 
-    my %args = $self->_get_form_data($r);
-
-    $self->_convert_to_get($r, \%args) if $r->method eq 'POST';
+    $self->_convert_to_get($r) if $r->method eq 'POST';
 
     # There should be a PerlSetVar directive that gives us the URI of
     # the script to execute for the login form.
@@ -331,9 +319,32 @@ sub login_form {
     }
 
     #$r->log_error("Redirecting to $authen_script");
-    $r->custom_response(FORBIDDEN, $authen_script);
+    my $status = $self->login_form_status($r);
+    $status = FORBIDDEN unless defined $status;
 
-    return FORBIDDEN;
+    if ($status == OK) {
+        # custom_response doesn't work for OK, DONE, or DECLINED in apache 1.x
+        $r->internal_redirect($authen_script);
+    }
+    else {
+        $r->custom_response($status, $authen_script);
+    }
+
+    return $status;
+}
+
+sub login_form_status {
+    my ($self, $r) = @_;
+
+    my $ua = $r->headers_in->get('User-Agent')
+        or return FORBIDDEN;
+
+    if (Apache::AuthCookie::Util::understands_forbidden_response($ua)) {
+        return FORBIDDEN;
+    }
+    else {
+        return OK;
+    }
 }
 
 sub satisfy_is_valid {
@@ -375,9 +386,8 @@ sub authorize ($$) {
     my $reqs_arr = $r->requires or return DECLINED;
 
     my $user = $r->connection->user;
-    unless ($user) {
-
-        # user is either undef or =0 which means the authentication failed
+    if ($user->is_blank) {
+        # authentication failed
         $r->log_reason("No user authenticated", $r->uri);
         return FORBIDDEN;
     }
@@ -534,8 +544,6 @@ sub get_cookie_path {
 
 1;
 
-
-
 =pod
 
 =head1 NAME
@@ -544,7 +552,7 @@ Apache::AuthCookie - Perl Authentication and Authorization via cookies
 
 =head1 VERSION
 
-version 3.18
+version 3.20
 
 =head1 SYNOPSIS
 
@@ -902,6 +910,17 @@ specified with the C<PerlSetVar WhatEverLoginScript> configuration
 directive. You can overwrite this method to provide your own
 mechanism.
 
+=item * login_form_status($r)
+
+This method returns the HTTP status code that will be returned with the login
+form response.  The default behaviour is to return FORBIDDEN, except for some
+known browsers which ignore HTML content for FORBIDDEN responses (e.g.:
+SymbianOS).  You can override this method to return custom codes.
+
+Note that FORBIDDEN is the most correct code to return as the given request was
+not authorized to view the requested page.  You should only change this if
+FORBIDDEN does not work.
+
 =item * logout()
 
 This is simply a convenience method that unsets the session key for
@@ -1068,9 +1087,19 @@ versions 2.x were written by Ken Williams <ken@forum.swarthmore.edu>
 
 L<perl(1)>, L<mod_perl(1)>, L<Apache(1)>.
 
+=head1 SOURCE
+
+The development version is on github at L<http://github.com/mschout/apache-authcookie>
+and may be cloned from L<git://github.com/mschout/apache-authcookie.git>
+
+=head1 BUGS
+
+Please report any bugs or feature requests to bug-apache-authcookie@rt.cpan.org or through the web interface at:
+ http://rt.cpan.org/Public/Dist/Display.html?Name=Apache-AuthCookie
+
 =head1 AUTHOR
 
-  Michael Schout <mschout@cpan.org>
+Michael Schout <mschout@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
@@ -1079,14 +1108,8 @@ This software is copyright (c) 2000 by Ken Williams.
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
 
-=head1 BUGS
-
-Please report any bugs or feature requests to bug-apache-authcookie@rt.cpan.org or through the web interface at:
- http://rt.cpan.org/Public/Dist/Display.html?Name=Apache-AuthCookie
-
 =cut
 
-
 __END__
 
 
