@@ -1,6 +1,6 @@
 package Apache2::AuthCookie;
 {
-  $Apache2::AuthCookie::VERSION = '3.19';
+  $Apache2::AuthCookie::VERSION = '3.20';
 }
 
 # ABSTRACT: Perl Authentication and Authorization via cookies
@@ -188,7 +188,7 @@ sub login {
     unless ($ses_key) {
         $r->server->log_error("Bad credentials") if $debug >= 2;
         $r->subprocess_env('AuthCookieReason', 'bad_credentials');
-        $r->uri($params->param('destination'));
+        $r->uri($self->untaint_destination($params->param('destination')));
         return $auth_type->login_form($r);
     }
 
@@ -548,7 +548,7 @@ Apache2::AuthCookie - Perl Authentication and Authorization via cookies
 
 =head1 VERSION
 
-version 3.19
+version 3.20
 
 =head1 SYNOPSIS
 
